//
//  SoftTimers Library - v1.1 - 06/05/2016
//  Copyright (C) 2016 Antoine Beauchamp
//  The code & updates for the library can be found on http://end2endzone.com
//
// AUTHOR/LICENSE:
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 3.0 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License (LGPL-3.0) for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// DISCLAIMER:
//  This software is furnished "as is", without technical support, and with no 
//  warranty, express or implied, as to its usefulness for any purpose.
//
// PURPOSE:
//  This Timer library regroup basic timer functionality into a single class.
//  The usual way to get the same functionnality is to create multiple
//  variables for a single timer. This way is hard to maintain when you need 
//  multiple timers running at the same time. SoftTimer classes are keep 
//  "simple & stupid". No software interrupts. Non-blocking. Each timer must
//  be polled within the loop().
//
// USAGE:
//  Call setTimeOutTime() to setup the non-blocking Timer. Then call 
//  reset() to restart the internal counter. Within the loop(), use
//  hasTimedOut() to know if the timer has triggered. Call getElapsedTime()
//  to get the absolute elapsed time since the last reset().
//  Units:
//    SoftTimerMillis() has all call in millisecond
//    SoftTimerMicros() has all call in microsecond
//
// HISTORY:
// 02/06/2016 v1.0 - Initial release.
// 06/05/2016 v1.1 - Fixed a bug where uninitialized timeout time would return incorrect results.
//                 - Added support for using an external counting function. This also allows the same class (SoftTimer)
//                   to support both milliseconds and microseconds.
//                 - Implemented progress and looping support. See getProgress(), getLoopCount() and getLoopProgress()
//                 - Implemented support for remaining time. See getRemainingTime()
//                 - Created new unit tests using win32Arduino library available at:
//                   http://www.end2endzone.com/win32arduino-a-win32-library-that-implements-most-arduino-functions/
//                 - Added many new examples to show possible use of the library.
//

#include "Arduino.h"
#include "SoftTimers.h"

SoftTimer::SoftTimer()
{
  mTimeOutTime = (uint32_t)0xFFFFFFFF; //maximum, never times out
  mCntFuncPtr = &millis;
  reset();
}

SoftTimer::SoftTimer(CounterFunctionPointer iCntFuncPtr)
{
  mTimeOutTime = (uint32_t)0xFFFFFFFF; //maximum
  mCntFuncPtr = iCntFuncPtr;
  reset();
}

SoftTimer::~SoftTimer()
{
}

void SoftTimer::reset()
{
  mStartTime = mCntFuncPtr();
}

uint32_t SoftTimer::getElapsedTime()
{
  //note: this code is actually wrap around safe.
  //tested on windows platform and Arduino Nano v3.
  uint32_t now = mCntFuncPtr();
  uint32_t elapsedTime = now - mStartTime;
  return elapsedTime;
}

uint32_t SoftTimer::getRemainingTime()
{
  //note: this code is actually wrap around safe.
  //tested on windows platform and Arduino Nano v3.
  uint32_t now = mCntFuncPtr();
  uint32_t elapsedTime = now - mStartTime;
  if (elapsedTime > mTimeOutTime)
    return 0;
  uint32_t remaining = mTimeOutTime - elapsedTime;
  return remaining;
}

void SoftTimer::setTimeOutTime(uint32_t iTimeOutTime)
{
  mTimeOutTime = iTimeOutTime;
}

uint32_t SoftTimer::getTimeOutTime()
{
  return mTimeOutTime;
}

bool SoftTimer::hasTimedOut()
{
  //update time value
  uint32_t elapsedTime = getElapsedTime();
  if (elapsedTime >= mTimeOutTime)
    return true;
  return false;
}

double SoftTimer::getProgress()
{
  uint32_t elapsed = getElapsedTime();

  //warning: elapsed time can actually be greater than time out time.
  if (elapsed >= mTimeOutTime)
    return 1.0;

  //when elapsed is < than time out time.
  return double(elapsed)/double(mTimeOutTime);
}

uint32_t SoftTimer::getLoopCount()
{
  uint32_t elapsed = getElapsedTime();

  if (elapsed == 0)
    return 0;

  //note that elapsed time is decreased of 1 unit
  //this allows to have a count of 0 when exactly 
  //timing out.
  uint32_t count = (elapsed-1)/mTimeOutTime;
  return count;
}

double SoftTimer::getLoopProgress()
{
  uint32_t elapsed = getElapsedTime();
  uint32_t progress = elapsed % (mTimeOutTime);
  return double(progress)/double(mTimeOutTime);
}
