//
//  SoftTimers Library - v1.1 - 06/05/2016
//  Copyright (C) 2016 Antoine Beauchamp
//  The code & updates for the library can be found on http://end2endzone.com
//
// AUTHOR/LICENSE:
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 3.0 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License (LGPL-3.0) for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// DISCLAIMER:
//  This software is furnished "as is", without technical support, and with no 
//  warranty, express or implied, as to its usefulness for any purpose.
//
// PURPOSE:
//  This Timer library regroup basic timer functionality into a single class.
//  The usual way to get the same functionnality is to create multiple
//  variables for a single timer. This way is hard to maintain when you need 
//  multiple timers running at the same time. SoftTimer classes are keep 
//  "simple & stupid". No software interrupts. Non-blocking. Each timer must
//  be polled within the loop().
//
// USAGE:
//  Call setTimeOutTime() to setup the non-blocking Timer. Then call 
//  reset() to restart the internal counter. Within the loop(), use
//  hasTimedOut() to know if the timer has triggered. Call getElapsedTime()
//  to get the absolute elapsed time since the last reset().
//  Units:
//    SoftTimerMillis() has all call in millisecond
//    SoftTimerMicros() has all call in microsecond
//
// HISTORY:
// 02/06/2016 v1.0 - Initial release.
// 06/05/2016 v1.1 - Fixed a bug where uninitialized timeout time would return incorrect results.
//                 - Added support for using an external counting function. This also allows the same class (SoftTimer)
//                   to support both milliseconds and microseconds.
//                 - Implemented progress and looping support. See getProgress(), getLoopCount() and getLoopProgress()
//                 - Implemented support for remaining time. See getRemainingTime()
//                 - Created new unit tests using win32Arduino library available at:
//                   http://www.end2endzone.com/win32arduino-a-win32-library-that-implements-most-arduino-functions/
//                 - Added many new examples to show possible use of the library.
//

#ifndef SoftTimers_h
#define SoftTimers_h

#include "Arduino.h"

class SoftTimer
{
public:
  /****************************************************************************
   * Description:
   *   Defines a generic 32 bits counting function.
   ****************************************************************************/
  typedef uint32_t (*CounterFunctionPointer)();

  /****************************************************************************
   * Description:
   *   Creates a SoftTimer instance which units is milliseconds.
   ****************************************************************************/
  SoftTimer();

  /****************************************************************************
   * Description:
   *   Creates a SoftTimer instance with custom unit.
   * Parameters:
   *   iCntFuncPtr: Pointer to a counting function.
   *                Use &millis to get a SoftTimer configured in milliseconds
   *                Use &micros to get a SoftTimer configured in microseconds
   ****************************************************************************/
  SoftTimer(CounterFunctionPointer iCntFuncPtr);

  /****************************************************************************
   * Description:
   *   Destructor
   ****************************************************************************/
  ~SoftTimer();

  /****************************************************************************
   * Description:
   *   reset() resets the internal counter of the Timer.
   ****************************************************************************/
  void reset();

  /****************************************************************************
   * Description:
   *   getElapsedTime() returns the elapsed time since the last reset()
   *   and now.
   ****************************************************************************/
  uint32_t getElapsedTime();

  /****************************************************************************
   * Description:
   *   getRemainingTime() returns the remaining time until the timer times out.
   *   Useful if one needs to show feedback to user.
   ****************************************************************************/
  uint32_t getRemainingTime();

  /****************************************************************************
   * Description:
   *   setTimeOutTime() defines the time out time at which hasTimedOut()
   *   will return true.
   * Parameters:
   *   iTimeOutTime: The given time out time in millisecond.
   ****************************************************************************/
  void setTimeOutTime(uint32_t iTimeOutTime);

  /****************************************************************************
   * Description:
   *   getTimeOutTime() returns the time out time at which hasTimedOut()
   *   will return true.
   ****************************************************************************/
  uint32_t getTimeOutTime();

  /****************************************************************************
   * Description:
   *   hasTimedOut() returns true if the elapsed time since reset() is
   *   greater or equal than the given time out time.
   ****************************************************************************/
  bool hasTimedOut();

  /****************************************************************************
   * Description:
   *   getProgress() returns the current timer progress in percentage
   *   (within [0, 1.0]) based on start time up to time out time.
   ****************************************************************************/
  double getProgress();
  
  /****************************************************************************
   * Description:
   *   getLoopProgress() returns the progress in percentage (within [0, 1.0])
   *   based on start time up to time out time as if the timer would
   *   automatically reset() when timing out.
   *   Note that a progress of 0% is the same as a progress of 100%.
   ****************************************************************************/
  double getLoopProgress();

  /****************************************************************************
   * Description:
   *   getLoopCount() returns the number of times the timer would have expired
   *   if the timer would automatically reset() when timing out.
   *   Useful for calculating pins that should toggle when timer expires.
   ****************************************************************************/
  uint32_t getLoopCount();

protected:
  CounterFunctionPointer mCntFuncPtr;
  volatile uint32_t mStartTime; 
  uint32_t mTimeOutTime;
};


//define legacy classes for backward compatibility
class SoftTimerMillis : public SoftTimer
{
public:
  SoftTimerMillis() : SoftTimer(&millis)
  {
  }
  ~SoftTimerMillis() {}
};

class SoftTimerMicros : public SoftTimer
{
public:
  SoftTimerMicros() : SoftTimer(&micros)
  {
  }
  ~SoftTimerMicros() {}
};

#endif
